@@ -1,17 +1,20 @@
-import asyncio
 import asyncio.coroutines
 import functools
+import inspect
 import os
 import sys
 import threading
+import warnings
+import weakref
 from concurrent.futures import Future, ThreadPoolExecutor
+from typing import Any, Callable, Dict, Optional, Union
 
 from .current_thread_executor import CurrentThreadExecutor
 from .local import Local
 
-try:
-    import contextvars  # Python 3.7+ only.
-except ImportError:
+if sys.version_info >= (3, 7):
+    import contextvars
+else:
     contextvars = None
 
 
@@ -26,6 +29,74 @@ def _restore_context(context):
             cvar.set(context.get(cvar))
 
 
+def _iscoroutinefunction_or_partial(func: Any) -> bool:
+    # Python < 3.8 does not correctly determine partially wrapped
+    # coroutine functions are coroutine functions, hence the need for
+    # this to exist. Code taken from CPython.
+    if sys.version_info >= (3, 8):
+        return asyncio.iscoroutinefunction(func)
+    else:
+        while inspect.ismethod(func):
+            func = func.__func__
+        while isinstance(func, functools.partial):
+            func = func.func
+
+        return asyncio.iscoroutinefunction(func)
+
+
+class ThreadSensitiveContext:
+    """Async context manager to manage context for thread sensitive mode
+
+    This context manager controls which thread pool executor is used when in
+    thread sensitive mode. By default, a single thread pool executor is shared
+    within a process.
+
+    In Python 3.7+, the ThreadSensitiveContext() context manager may be used to
+    specify a thread pool per context.
+
+    In Python 3.6, usage of this context manager has no effect.
+
+    This context manager is re-entrant, so only the outer-most call to
+    ThreadSensitiveContext will set the context.
+
+    Usage:
+
+    >>> import time
+    >>> async with ThreadSensitiveContext():
+    ...     await sync_to_async(time.sleep, 1)()
+    """
+
+    def __init__(self):
+        self.token = None
+
+    if contextvars:
+
+        async def __aenter__(self):
+            try:
+                SyncToAsync.thread_sensitive_context.get()
+            except LookupError:
+                self.token = SyncToAsync.thread_sensitive_context.set(self)
+
+            return self
+
+        async def __aexit__(self, exc, value, tb):
+            if not self.token:
+                return
+
+            executor = SyncToAsync.context_to_thread_executor.pop(self, None)
+            if executor:
+                executor.shutdown()
+            SyncToAsync.thread_sensitive_context.reset(self.token)
+
+    else:
+
+        async def __aenter__(self):
+            return self
+
+        async def __aexit__(self, exc, value, tb):
+            pass
+
+
 class AsyncToSync:
     """
     Utility class which turns an awaitable that only works on the thread with
@@ -40,13 +111,17 @@ class AsyncToSync:
     """
 
     # Maps launched Tasks to the threads that launched them (for locals impl)
-    launch_map = {}
+    launch_map: "Dict[asyncio.Task[object], threading.Thread]" = {}
 
     # Keeps track of which CurrentThreadExecutor to use. This uses an asgiref
     # Local, not a threadlocal, so that tasks can work out what their parent used.
     executors = Local()
 
     def __init__(self, awaitable, force_new_loop=False):
+        if not callable(awaitable) or not _iscoroutinefunction_or_partial(awaitable):
+            # Python does not have very reliable detection of async functions
+            # (lots of false negatives) so this is just a warning.
+            warnings.warn("async_to_sync was passed a non-async-marked callable")
         self.awaitable = awaitable
         try:
             self.__self__ = self.awaitable.__self__
@@ -206,11 +281,11 @@ class AsyncToSync:
             if exc_info[1]:
                 try:
                     raise exc_info[1]
-                except Exception:
+                except BaseException:
                     result = await self.awaitable(*args, **kwargs)
             else:
                 result = await self.awaitable(*args, **kwargs)
-        except Exception as e:
+        except BaseException as e:
             call_result.set_exception(e)
         else:
             call_result.set_result(result)
@@ -237,6 +312,10 @@ class SyncToAsync:
     outermost), this will just be the main thread. This is achieved by idling
     with a CurrentThreadExecutor while AsyncToSync is blocking its sync parent,
     rather than just blocking.
+
+    If executor is passed in, that will be used instead of the loop's default executor.
+    In order to pass in an executor, thread_sensitive must be set to False, otherwise
+    a TypeError will be raised.
     """
 
     # If they've set ASGI_THREADS, update the default asyncio executor for now
@@ -247,7 +326,7 @@ class SyncToAsync:
         )
 
     # Maps launched threads to the coroutines that spawned them
-    launch_map = {}
+    launch_map: "Dict[threading.Thread, asyncio.Task[object]]" = {}
 
     # Storage for main event loop references
     threadlocal = threading.local()
@@ -255,13 +334,38 @@ class SyncToAsync:
     # Single-thread executor for thread-sensitive code
     single_thread_executor = ThreadPoolExecutor(max_workers=1)
 
-    def __init__(self, func, thread_sensitive=True):
+    # Maintain a contextvar for the current execution context. Optionally used
+    # for thread sensitive mode.
+    if sys.version_info >= (3, 7):
+        thread_sensitive_context: "contextvars.ContextVar[str]" = (
+            contextvars.ContextVar("thread_sensitive_context")
+        )
+    else:
+        thread_sensitive_context: None = None
+
+    # Maintaining a weak reference to the context ensures that thread pools are
+    # erased once the context goes out of scope. This terminates the thread pool.
+    context_to_thread_executor: "weakref.WeakKeyDictionary[object, ThreadPoolExecutor]" = (
+        weakref.WeakKeyDictionary()
+    )
+
+    def __init__(
+        self,
+        func: Callable[..., Any],
+        thread_sensitive: bool = True,
+        executor: Optional["ThreadPoolExecutor"] = None,
+    ) -> None:
+        if not callable(func) or _iscoroutinefunction_or_partial(func):
+            raise TypeError("sync_to_async can only be applied to sync functions.")
         self.func = func
         functools.update_wrapper(self, func)
         self._thread_sensitive = thread_sensitive
-        self._is_coroutine = asyncio.coroutines._is_coroutine
+        self._is_coroutine = asyncio.coroutines._is_coroutine  # type: ignore
+        if thread_sensitive and executor is not None:
+            raise TypeError("executor must not be set when thread_sensitive is True")
+        self._executor = executor
         try:
-            self.__self__ = func.__self__
+            self.__self__ = func.__self__  # type: ignore
         except AttributeError:
             pass
 
@@ -273,11 +377,26 @@ class SyncToAsync:
             if hasattr(AsyncToSync.executors, "current"):
                 # If we have a parent sync thread above somewhere, use that
                 executor = AsyncToSync.executors.current
+            elif self.thread_sensitive_context and self.thread_sensitive_context.get(
+                None
+            ):
+                # If we have a way of retrieving the current context, attempt
+                # to use a per-context thread pool executor
+                thread_sensitive_context = self.thread_sensitive_context.get()
+
+                if thread_sensitive_context in self.context_to_thread_executor:
+                    # Re-use thread executor in current context
+                    executor = self.context_to_thread_executor[thread_sensitive_context]
+                else:
+                    # Create new thread executor in current context
+                    executor = ThreadPoolExecutor(max_workers=1)
+                    self.context_to_thread_executor[thread_sensitive_context] = executor
             else:
                 # Otherwise, we run it in a fixed single thread
                 executor = self.single_thread_executor
         else:
-            executor = None  # Use default
+            # Use the passed in executor, or the loop's default if it is None
+            executor = self._executor
 
         if contextvars is not None:
             context = contextvars.copy_context()
@@ -298,7 +417,7 @@ class SyncToAsync:
                 sys.exc_info(),
                 func,
                 *args,
-                **kwargs
+                **kwargs,
             ),
         )
         ret = await asyncio.wait_for(future, timeout=None)
@@ -337,7 +456,7 @@ class SyncToAsync:
             if exc_info[1]:
                 try:
                     raise exc_info[1]
-                except Exception:
+                except BaseException:
                     return func(*args, **kwargs)
             else:
                 return func(*args, **kwargs)
@@ -369,7 +488,19 @@ class SyncToAsync:
 async_to_sync = AsyncToSync
 
 
-def sync_to_async(func=None, thread_sensitive=True):
+def sync_to_async(
+    func: Optional[Callable[..., Any]] = None,
+    thread_sensitive: bool = True,
+    executor: Optional["ThreadPoolExecutor"] = None,
+) -> Union[SyncToAsync, Callable[[Callable[..., Any]], SyncToAsync]]:
     if func is None:
-        return lambda f: SyncToAsync(f, thread_sensitive=thread_sensitive)
-    return SyncToAsync(func, thread_sensitive=thread_sensitive)
+        return lambda f: SyncToAsync(
+            f,
+            thread_sensitive=thread_sensitive,
+            executor=executor,
+        )
+    return SyncToAsync(
+        func,
+        thread_sensitive=thread_sensitive,
+        executor=executor,
+    )
